/* hf_io.c   do input/output on heightfield data  JPB july 22 1995 */

#include <stdio.h>                      /* fopen() sprintf() etc. */
#include <stdlib.h>                     /* strtod() */
#include <math.h>                       /* math stuff */
#include <string.h>                     /* strcpy() */
#include <ctype.h>                      /* isspace(), isdigit() -- RRB */
#include "hl.h"
#include "hf_io.h"

int readn_hf(hfield *hfp, char *fname_in)
{
int inlen, rval, itype;
char ext1[5];

  inlen = strlen(fname_in);
  strncpy(ext1,(char *)(&fname_in[inlen-4]),5);

  if (ext1 == NULL) { 
    fprintf(stderr,"filenames must have type extension, eg. topo.gif\n");
    return(1);
  }

  if       (strcmp(ext1,".gif")==0) itype = GIF;  
   else if (strcmp(ext1,".pot")==0) itype = POT;
   else if (strcmp(ext1,".tga")==0) itype = TGA;
   else if (strcmp(ext1,".pgm")==0) itype = PGM;
   else if (strcmp(ext1,".pg8")==0) itype = PG8;
   else if (strcmp(ext1,".mat")==0) itype = MAT;
   else if (strcmp(ext1,".oct")==0) itype = OCT;
   else if (strcmp(ext1,".png")==0) itype = PNG;
   else { 
     itype = PGM;       /* default type */
     fprintf(stderr,"Warning: %s has a missing or unknown file extension\n",
	     fname_in);
   }
					/* read in heightfield */
   rval = read_hf(hfp, fname_in, itype);
   return(rval);
} /* end readn_hf */

int writen_hf(hfield hf, char *fname_out)
{
int outlen, rval, itype;
char ext1[6];

  outlen = strlen(fname_out);
  strncpy(ext1,(char *)(&(fname_out[outlen-4])),5);

  if (ext1 == NULL) { 
    fprintf(stderr,"filenames must have type extension, eg. topo.gif\n");
    return(1);
  }

  if       (strcmp(ext1,".gif")==0) itype = GIF;  
   else if (strcmp(ext1,".pot")==0) itype = POT;
   else if (strcmp(ext1,".tga")==0) itype = TGA;
   else if (strcmp(ext1,".raw")==0) itype = RAW;
   else if (strcmp(ext1,".pgm")==0) itype = PGM;
   else if (strcmp(ext1,".pg8")==0) itype = PG8;
   else if (strcmp(ext1,".mat")==0) itype = MAT;
   else if (strcmp(ext1,".oct")==0) itype = OCT;
   else if (strcmp(ext1,".png")==0) itype = PNG;
   else { 
     fprintf(stderr,"Error: %s has a missing or unknown file extension %s\n",
	     fname_out,ext1);
     return(1);
   }
			/* write out heightfield of type itype */       
   rval = write_hf(hf, fname_out, itype);
   return(rval);
} /* end writen_hf */

      /* print out hf array to PGM ascii */
int hf_write(hfield hf, char *fname)
{
  int xsize, ysize;
  int cnt;              /* count of #s printed */
  int ix,iy;
  long idx;
  FILE *fpout;
  int rval;
  int maxheight = 65535;        /* maximum value of array */
						/* open input file */
  if ((fpout = fopen(fname,"w")) == NULL)
   { perror("Could not open output file ");
     return(1);
   }

  xsize = hf.xsize;
  ysize = hf.ysize;
  fprintf(fpout,"P2\n");
  fprintf(fpout,"%d %d\n",xsize,ysize);
  fprintf(fpout,"%d\n",maxheight);

  idx = 0;
  cnt = 0;
   for (iy = 0; iy<ysize; iy++) {
    for (ix = 0; ix<xsize; ix++) {
	fprintf(fpout,"%d ",(int) (maxheight*hf.a[idx++]));
	if (++cnt > 10) { fprintf(fpout,"\n"); cnt=0; }
    }
    if (cnt != 0) { fprintf(fpout,"\n"); cnt = 0;}
   }
   fclose(fpout);         /* end of file */
   rval = 0;
   return(rval);
} /* hf_write() */

/* oct_write()  --  print out hf array to Octave ascii fmt */

int oct_write(hfield hf, char *fname, char *mname)  
{
  int xsize, ysize;
  int ix,iy;
  long idx;
  FILE *fpout;
  int rval;
						/* open input file */
  if ((fpout = fopen(fname,"w")) == NULL)
   { perror("Could not open output file ");
     return(1);
   }

  xsize = hf.xsize;
  ysize = hf.ysize;
  fprintf(fpout,"# name: %s\n",mname);
  fprintf(fpout,"# type: matrix\n");
  fprintf(fpout,"# rows: %d\n",ysize);
  fprintf(fpout,"# columns: %d\n",xsize);
  
  idx = 0;
   for (iy = 0; iy<ysize; iy++) {
    for (ix = 0; ix<xsize; ix++) {
	fprintf(fpout,"%e ",hf.a[idx++]);
    }
     fprintf(fpout,"\n");
   }
   fclose(fpout);
   rval = 0;
   return(rval);
} /* oct_write */

		/* write hf array to PGM binary */
int hf_writeb(hfield hf, char *fname)
{
  int xsize, ysize;
  int ix,iy;
  long idx;
  FILE *fpout;
  int rval;
  int maxheight = 255;        /* maximum value of array */
						/* open input file */
  if ((fpout = fopen(fname,"wb")) == NULL)
   { perror("Could not open output file ");
     return(1);
   }

  xsize = hf.xsize;
  ysize = hf.ysize;
  fprintf(fpout,"P5\n");
  fprintf(fpout,"%d %d\n",xsize,ysize);
  fprintf(fpout,"%d\n",maxheight);

  idx = 0;
   for (iy = 0; iy<ysize; iy++) {
    for (ix = 0; ix<xsize; ix++) {
	fprintf(fpout,"%c",(BYTE)(255*hf.a[idx++]));
    }
   }
   fclose(fpout);
   rval = 0;
   return(rval);
} /* end hf_writeb() */

/* -----------------------------------------------------------   */
/* hf_read -- read in pgm data from file <fname> into hf array,  */
/* allocating memory to fit a (xsize x ysize) array.             */
/* readin PGM binary or ascii, or if neither, call read_mfile()  */
/* -----------------------------------------------------------   */

int hf_read(hfield *hfp, char *fname)

{
  hfield hf;                            /* working heightfield */
  int xsize,ysize;
  unsigned long npoints;            /* # points that should be in file */
  unsigned long points_read;        /* # points read from hf datafile */
  char sbuf[512],                       /* buffer for file input */
	dummy[2];                       /* used for \n and \0   */
  char *junk;                           /* unconverted part of a number */
  double a, maxval;
  FILE *fpin;                           /* input file pointer */
  BYTE fbin;                            /* PGM binary format flag */
  BYTE oneb;                            /* one byte from file */
  int rval;
						/* open input file */
  if ((fpin = fopen(fname,"rb")) == NULL)
   { perror("Could not open input file ");
     return(1);
   }

  if (fread(sbuf,sizeof(char),2,fpin) != 2)
    {   perror("Could not read file"); return(1); }
  else sbuf[2] = 0x00;

  fbin = 0;                             /* assume PGM ascii */
  if (strcmp(sbuf,"P5") == 0) fbin = 1; /* PGM binary format */
  if (strcmp(sbuf,"GI") == 0) {
    fclose(fpin);
    rval = gread1(hfp,fname);        /* read as GIF format file */
  } else if ((strcmp(sbuf,"P5") != 0) && (strcmp(sbuf,"P2") != 0)) {
					      /* not PGM Ascii */
	  fclose(fpin);   /* close file to reopen as binary */
	  rval = read_mfile(hfp,fname);
  } else {  /* read PGM ASCII or binary file */

  while ((fscanf(fpin,"%s",sbuf) != EOF) && sbuf[0] == '#') {
/*        printf("comment:"); */
	fscanf(fpin,"%[^\n]",sbuf);  /* read comment beyond '#' */
	fscanf(fpin,"%[\n]",dummy);  /* read newline */
/*        printf("%s\n",sbuf);    */        /* display comment */
	}

  xsize = atoi(sbuf);                   /* store xsize of array */
  fscanf(fpin,"%s",sbuf);               /* read ysize of array */
  ysize = atoi(sbuf);
  fscanf(fpin,"%s\n",sbuf);               /* read maxval of array */
  maxval = strtod(sbuf, &junk);         /* store maxval. could throw away. */

  if (h_newr(&hf,xsize,ysize,"fname")) return(1);

  points_read = 0;                     /* haven't read any data points yet */
  npoints = ((unsigned long int)xsize*ysize); /* # of points to read in */

  if (fbin) {  /* read PGM binary */
   while ((fread(&oneb,sizeof(BYTE),1,fpin) == 1) && (points_read<npoints)) {
      hf.a[points_read++] = (PTYPE) oneb / 255;
   }
  } else { /* read PGM ASCII */
      while ((fscanf(fpin,"%s",sbuf) != EOF) && (points_read < npoints) ) {
	a = strtod( sbuf, &junk);
	hf.a[points_read++] = (PTYPE) a / maxval;
      }
  } /* end else */

  printf("%lu elements read.\n",points_read);
  if (points_read != npoints) {
	printf("Warning: %lu points missing.\n",npoints-points_read);
   }
  rval = fclose( fpin );
  *hfp = hf;            /* assign the HF to return value pointer */
  } /* end else (read as PGM ascii/binary) */

  return(rval);
} /* end hf_read */

/* -------------------------------------------------------------- */
/* read_mfile --  read in data matrix in Matlab Level 1.0 format  */
/*               and put data in header "x" and data array "hf"   */
/* -------------------------------------------------------------- */

int read_mfile(hfield *hfp, char *fname )
{
  hfield hf;                    /* heightfield structure */
  int xsize,ysize;              /* dimensions */
  int cflag;                    /* complex type flag 0=real, 1=complex */
  char pname[80]="dat1\x00";           /* data matrix name, null-terminated*/
  Fmatrix hd;                           /* Level 1.0 MAT-file header */

  int a,b,c,i,j;
  int etype;
  int mtype,prec,rem,numtype;
  double tdbl;
  float tflt;
  FILE *fpin;

   /* --------- read in Matlab-format (Level 1.0) file ------- */

   etype = (NUM_FMT*1000) + (PREC*10);    /* expected data type flag */

   if ((fpin = fopen(fname,"rb")) == NULL) {
    perror("could not open file.\n");
    return(1);
   }

   a = fread(&hd, sizeof(Fmatrix),(size_t)1, fpin);   /* read header */
    if (a != 1) {
      perror("Problem reading matlab file header.\n");
      return(1);
    }
   numtype = (hd.type % 10000)/1000;
   prec = (hd.type % 100)/10;
   mtype = (hd.type % 10);
   rem = (hd.type - (1000*numtype + 10*prec + mtype));
   cflag = hd.imagf;
  /*  printf("Data type flag: M%d P%d T%d R%d CPLX%d\n",
       numtype,prec,mtype,rem,cflag); */
   if ( (numtype!=NUM_FMT) || ((prec!=0)&&(prec!=1)) || (mtype!=0) )  {
     perror("Unsupported file format type\n");
     return(1);
   }
   b = fread(pname, sizeof(char), (size_t)hd.namelen, fpin); /* read mx name */
   xsize = hd.ncols;
   ysize = hd.mrows;
   if (cflag) {
     if (h_newc(&hf,xsize,ysize,fname)) return(1);  /* alloc new HF */
   } else {
     if (h_newr(&hf,xsize,ysize,fname)) return(1);  /* alloc new HF */
   }
   
   if (prec==1) {               /* read in floats 1 by 1 */
     for (i=0;i<xsize;i++) { for (j=0;j<ysize;j++) {   /* columns first */
       c = fread(&tflt, sizeof(float), 1, fpin);
       if (c != 1) {perror("Trouble reading data from file.\n"); return(1); }
       El(hf.a,i,j) = (PTYPE) tflt; } /* end for j */ } /* end for i */
     if (cflag) {
      for (i=0;i<xsize;i++) { for (j=0;j<ysize;j++) {   /* columns first */
       c = fread(&tflt, sizeof(float), 1, fpin);
       if (c != 1) {perror("Trouble reading data from file.\n"); return(1); }
       Im(hf.a,i,j) = (PTYPE) tflt; } /* end for j */ } /* end for i */
      }

   } else if (prec==0) {     /* read doubles 1 by 1 and convert */

     for (i=0;i<xsize;i++) {for (j=0;j<ysize;j++) {   /* columns first */
       c = fread(&tdbl, sizeof(double), 1, fpin);
       if (c != 1) {perror("Trouble reading data from file.\n");return(1);}
       El(hf.a,i,j) = (PTYPE) tdbl; } /* end for j */ } /* end for i */
     if (cflag) {
      for (i=0;i<xsize;i++) {for (j=0;j<ysize;j++) {   /* columns first */
       c = fread(&tdbl, sizeof(double), 1, fpin);
       if (c != 1) {perror("Trouble reading data from file.\n");return(1);}
       Im(hf.a,i,j) = (PTYPE) tdbl; } /* end for j */ } /* end for i */
     }
   } else {
	perror("Unsupported floating-point type\n");
	return(1);
   } /* end else */

   *hfp = hf;
   fclose( fpin );
   return(0);
} /* end read_mfile */


/* --------------------------------------------------------------  */
/* write_mfile -- write out data matrix in Matlab Level 1.0 format */
/*               creates data header fhdr from xsize, ysize and    */
/*               writes data array 'hf'                            */
/* --------------------------------------------------------------  */

int write_mfile(hfield hf, char *fname, char *pname)
{
FILE *fpout;
int xsize, ysize;
Fmatrix fhdr;
int rv,mn,i,j,c;

   xsize = hf.xsize;
   ysize = hf.ysize;
   fpout = fopen(fname,"wb");
   if (fpout == NULL)
    { perror("Error opening file.\n");
      return(1);
    }
   fhdr.type = NUM_FMT*1000 + PREC*10;
   fhdr.ncols = xsize;
   fhdr.mrows = ysize;
   fhdr.imagf = hf.c;
   fhdr.namelen = strlen(pname) + 1;

   rv = fwrite(&fhdr, sizeof(Fmatrix),(size_t)1, fpout);          /* write header */
   if (rv != 1) {
     perror("Error writing file.\n");
     return(1);
   }
   rv = fwrite(pname, sizeof(char), (size_t)fhdr.namelen, fpout);   /* write mx name */
   if (rv != fhdr.namelen) {
     perror("Error writing file.\n");
     return(1);
   }
   mn = fhdr.mrows * fhdr.ncols;

   for (i=0;i<xsize;i++) { for (j=0;j<ysize;j++) {   /* columns first */
       c = fwrite(&El(hf.a,i,j), sizeof(PTYPE), 1, fpout);
       if (c != 1) {perror("Trouble writing data to file.\n");return(1);}
      } /* end for j */   } /* end for i */
   if (hf.c) {
     for (i=0;i<xsize;i++) { for (j=0;j<ysize;j++) {   /* columns first */
       c = fwrite(&Im(hf.a,i,j), sizeof(PTYPE), 1, fpout);
       if (c != 1) {perror("Trouble writing data to file.\n");return(1);}
      } /* end for j */   } /* end for i */
   }

  if (fclose( fpout ) == EOF) {perror("Error closing file.\n");return(1);}

  return(0);
} /* end write_mfile */

/* ------------------------------------------------------------------- 
  read_hf() --  read in a heightfield in GIF|POT|TGA|MAT|etc format
		 fname_in is filename string, itype is format type number
		 return dat in hfp, dimensions in (xsp, ysp)
  --------------------------------------------------------------------
*/

int read_hf(hfield *hfp, char *fname_in, int itype)
{
int rval; 

   if (itype == POT) {
    rval = pot_read(hfp, fname_in);
   } else if (itype == TGA) {
    rval = tga_read(hfp, fname_in);
#ifdef HAS_PNG
   } else if (itype == PNG) {
    rval = png_read(hfp, fname_in);
#else
   } else if (itype == PNG) {
    printf("Sorry, PNG support not compiled in.\n");
    rval = 1;
#endif
   } 
   else if ( itype == OCT )
   {
     rval = oct_read( hfp, fname_in );
   } else { 
       rval = hf_read(hfp, fname_in);  
     }
   return(rval);
} /* read_hf() */

/* ------------------------------------------------------------------- 
  write_hf() --  write out a heightfield in GIF|POT|TGA|MAT|etc format
		 fname_out is filename string, otype is format type number
		 hfield dat in hf, dimensions in (xsize, ysize)
  --------------------------------------------------------------------
*/

int write_hf(hfield hf, char *fname_out, int otype)
{
 int rval;

 if (hf.c && otype != MAT) {
   printf("Error: can only save complex matrix as .MAT\n");
   return(1);
 }

   if (otype == GIF) {
     printf("Writing GIF file...\n");
     rval = gwrite(hf, fname_out);
   } else if (otype == PGM) {
       printf("Writing PGM ascii...\n");
       rval = hf_write(hf, fname_out);  /* save output file */
   } else if (otype == PG8) {
     printf("Writing PGM binary...\n");
     rval = hf_writeb(hf, fname_out);  /* save output file */
   } else if (otype == POT) {
     printf("Writing POT file...\n");
     rval = pot_write(hf, fname_out);
   } else if (otype == TGA) {
     printf("Writing TGA file...\n");
     rval = tga_write(hf, fname_out);
   } else if (otype == RAW) {
     printf("Writing RAW file (%ld triangles)...\n",
         ((long)(hf.xsize-1) * (hf.ysize-1) * 2) );
     rval = raw_write(hf, fname_out);
   } else if (otype == MAT) {
     printf("Writing Matlab file...\n");
     rval = write_mfile(hf, fname_out,"topo");
   } else if (otype == OCT) {
     printf("Writing Octave file...\n");
     rval = oct_write(hf, fname_out,"topo");
#ifdef HAS_PNG
   } else if (otype == PNG) {
     printf("Writing PNG file...\n");
     rval = png_write(hf, fname_out);
#else
   } else if (otype == PNG) {
     printf("Sorry, PNG support not compiled in.\n");
     rval = 1;
#endif
   } else rval = 1;

   return(rval);
} /* end write_hf() */








/* --------------------------------------------------------------  */
/* oct_read  --  read in data in Octsave format                    */
/*        Modified by Robert R. Bertram July 21, 1997              */
/*          Can read in ".oct" format data                         */
/* --------------------------------------------------------------  */
/*  The following function reads in (almost) any type of generic   */
/* test image data including the OCT format.                       */
/* --------------------------------------------------------------  */


#define STRING_SIZE  16384         /* set (Max) string size to 16K */
#define ARRAY_SIZE    2048
#define NAME_SIZE       20

#define OCT_NAME      "name:"
#define OCT_TYPE      "type:"
#define OCT_ROWS      "rows:"
#define OCT_COLUMNS   "columns:"

#define META_DATA_CHARACTER '#'

#define NUM_LINES_READ   4

typedef enum OctInputFlags 
{ 
  OCT_None     = 0,
  OCT_Name     = 1,
  OCT_Type     = 2,
  OCT_Rows     = 4,
  OCT_Columns  = 8
} OCT_OIF;

int oct_read( hfield *hfp, char *fname )
{
    /* variables */
  int xsize, ysize;
  int i;
  long idx, elements_read;
  long int file_pos, file_pos_new;
  FILE *fpin;
  int oct_flag, break_flag;
  PTYPE array[ ARRAY_SIZE ];
  char in_str[ STRING_SIZE ],
       oct_name[ NAME_SIZE ],
       oct_type[ NAME_SIZE ];
  char * str_ptr, c ;
  char * str_name = OCT_NAME,
       * str_type = OCT_TYPE,
       * str_rows = OCT_ROWS,
       * str_col  = OCT_COLUMNS;
  
  /* executables */
						/* open input file */
  if ((fpin = fopen(fname,"r")) == NULL)
  { 
    perror( "Could not open [.oct] input file." );
    return(1);
  }
  
    /* initilize our data */
  xsize = 
    ysize = 0;
  idx =
    elements_read = 0;
  file_pos = file_pos_new = 0;             
  oct_flag = OCT_None;      /* received no info about OCT file yet */
  oct_name[0] = 
    oct_type[0] = 0;         /* strings of length zero */
  break_flag = 0;            /* flag to indicate a break in a loop */
  
  do 
  {
    
    file_pos = ftell( fpin );
      /* read in an entire line */
    if ( !fgets( in_str, ARRAY_SIZE, fpin ) )
    { 
      /* premature end of file (?)*/
      if( !xsize || !ysize || !elements_read )
      {  
        perror( "Unexpected end of file encountered while reading file." );
        return( 1 );
      }
        /* assume that we're done reading this image */
      return 0;
    }
    
    file_pos_new = ftell( fpin );
    
      /* find the first non-whitespace */
    str_ptr = in_str;
    while ( isspace( *str_ptr ) && *str_ptr )  str_ptr++;
    
      /* check to see if the file has image meta-data */
    if ( *str_ptr == META_DATA_CHARACTER )
    {
      /* This line contains information about the image */
      str_ptr++;
      while ( isspace( *str_ptr ) && *str_ptr )  str_ptr++;
      
        /* check for the image name */
      if ( !strncmp( str_ptr, str_name, strlen( str_name ) ) )
      {
        /* this has information about the image name */
        str_ptr += strlen( str_name );
        sscanf( str_ptr, "%s", oct_name );
        
        oct_flag |= OCT_Name;
        
        /* assume that there's no more information on this line */
        continue;
      }
        /* check for the image type */
      else if ( !strncmp( str_ptr, str_type, strlen( str_type ) ) )
      {
        /* this has information about the image type */
        str_ptr += strlen( str_type );
        sscanf( str_ptr, "%s", oct_type );
        
        oct_flag |= OCT_Type;
        
        /* assume that there's no more information on this line */
        continue;
      }
        /* check for the number of rows */
      if ( !strncmp( str_ptr, str_rows, strlen( str_rows ) ) )
      {
        /* this has information about the number of rows */
        str_ptr += strlen( str_rows );
        sscanf( str_ptr, "%d", &ysize );
        
        oct_flag |= OCT_Rows;
        
        /* assume that there's no more information on this line */
        continue;
      }
        /* check for the number of columns */
      else if ( !strncmp( str_ptr, str_col, strlen( str_col ) ) )
      {
        /* this has information about the image name */
        str_ptr += strlen( str_col );
        sscanf( str_ptr, "%d", &xsize );
        
        oct_flag |= OCT_Columns;
        
        /* assume that there's no more information on this line */
        continue;
      }
        /* do this on default */
      else
      {
        /* this line is either a comment(?) or the parse didn't */
        /* go correctly...  skip this line */
        continue;
      }
    }
    
    /* is our first character part of a number? */
    c = *str_ptr;
    if ( isdigit( c ) || c == '+' || c == '-' )
    {
      /* we're dealing with a number of some sort */
      
        /* do we know the dimensions of this image */
      if ( (oct_flag & OCT_Rows) &&  (oct_flag & OCT_Columns) )
      {
        /*  Yep, we have all that we need to start. */
        h_newr( hfp, xsize, ysize, oct_name );
        
          /* assume that the rest of the information in the file is */
          /* image data. */
          /* go back to the start of the line and start reading */
        fseek( fpin, file_pos, SEEK_SET);
        for ( i = elements_read; i < xsize * ysize; i++ )
        {
          if ( fscanf( fpin, "%e", &hfp -> a[i] ) != 1 )
          {
            perror( "Unexpected end of file encountered when reading [.oct]." );
            return ( 1 );
          }
        }
        break_flag = 1;  /* signify that we're done reading the file */
        break;
      }
      else if ( oct_flag & OCT_Columns )
      {
        /* we only know the number of columns */
          /* read in another row... */
        if ( !ysize )
        {
          /* haven't read in any data yet -- initilize the hf structure */
          h_newr( hfp, xsize, ++ysize, oct_name );
        }
        else
        {
            /* have already read in some image data, reallocate memory for */
            /* one more row */
          hfp -> ysize = ++ysize; 
          hfp -> a = (PTYPE *)realloc( hfp -> a,  
                   (size_t) xsize * ysize * sizeof( PTYPE ) );
          if ( !hfp -> a )
          {
            perror( "Memory allocation error when reading [.oct] file " );
            return 1;
          }
        }
        
          /* go back to the start of the line and start reading */
        fseek( fpin, file_pos, SEEK_SET);
        
          /* read in the next row */
        for ( i = 0; i < xsize; i++ )
          if ( fscanf( fpin, "%e", &hfp -> a[i + elements_read] ) != 1 )
          {
            perror( "Unexpected end of file encountered when reading [.oct]." );
            return ( 1 );
          }
        elements_read += xsize;  /* keep track of the number of elements read */
        /* done reading this line -- continue parsing the image file */

        fseek( fpin, file_pos_new, SEEK_SET);  /* go to the next line */
        
        continue;
      }
      else  /* if ( oct_flag & OCT_Rows ) */
      {
        /* we know the number of rows but not the number of columns */
        /* I don't know what type of situation would use this...    */
        
          /* conditions leading to this section of code:            */
          /*   o  have not read any image data yet                  */
          /*   o  xsize = 0                                         */
          /*   o  elements_read = 0                                 */
          
        while ( sscanf( str_ptr, "%e", &array[ xsize ] ) == 1 )
        {
          xsize++;  /* count the number of elements in this row */
          
            /* skip to next number in the string */
          while ( !isspace( *(++str_ptr) ) ) ;
          while ( isspace( *(++str_ptr ) ) ) ;
        }
        
        if ( !xsize )
        {
          perror( "Error reading:  No elements read in [.oct] file." );
          perror( "Do not understand the following:" );
          perror( str_ptr );
          return 1;
        }
        
        elements_read += xsize;

          /* allocate memory for this image */
        if ( h_newr( hfp, xsize, ++ysize, oct_name ) )
          return 1;  /* out of memory error */
          
        /* copy the array contents into hfp */
        memcpy( hfp -> a, array, (size_t) xsize * sizeof(PTYPE) );
        
        /* we now know the width of the image data */
        oct_flag |= OCT_Columns;
        continue;
      }
    }
    
    perror( "Do not understand the following:  " );
    perror( str_ptr );
  } while ( !break_flag );
      
  return 0;
}
 
